#include <map>
#include <ctime>
#include <sys/time.h>
#include <ANN.h>
#include "findMatch.h"
#include "detectFeatures.h"
#include "../emtv/StdAfx.h"
#include "../emtv/eig.h"

using namespace PMVS3;
using namespace std;
using namespace Patch;

CfindMatch::CfindMatch(void)
  : m_pos(*this), m_seed(*this), m_expand(*this), m_filter(*this), m_optim(*this) {
  m_debug = 0;
  m_maxsal = 0.0;
  m_addedpatches = 0;
  m_fSigmaScale = 1;    
  m_fEpsBoundary = 0.0075;
  R = NULL;
  Rp = NULL;
  outr = NULL;
  r = NULL;
  I = NULL;
}

CfindMatch::~CfindMatch() {
  pthread_rwlock_destroy(&m_lock);
  
  for (int image = 0; image < (int)m_imageLocks.size(); ++image)
    pthread_rwlock_destroy(&m_imageLocks[image]);
  for (int image = 0; image < (int)m_countLocks.size(); ++image)
    pthread_rwlock_destroy(&m_countLocks[image]);
}

void CfindMatch::updateThreshold(void) {
  m_nccThreshold -= 0.05f;
  m_nccThresholdBefore -= 0.05f;
  
  m_countThreshold1 = 2;
}

void CfindMatch::init(const Soption& option) {
  m_timages = option.m_timages;
  m_oimages = option.m_oimages;
  m_images.clear();
  m_images.insert(m_images.end(), m_timages.begin(), m_timages.end());
  m_images.insert(m_images.end(), m_oimages.begin(), m_oimages.end());
  
  m_tnum = (int)m_timages.size();
  m_num = (int)m_images.size();
  
  m_prefix = option.m_prefix;
  m_level = option.m_level;
  m_csize = option.m_csize;
  m_nccThreshold = option.m_threshold;
  m_wsize = option.m_wsize;
  m_minImageNumThreshold = option.m_minImageNum;
  m_CPU = option.m_CPU;
  m_setEdge = option.m_setEdge;
  m_sequenceThreshold = option.m_sequence;
  m_fSigmaScale = option.m_fSigmaScale;
  m_ffactor = option.m_factor;

  m_junit = 100;
  // This initialization does not matter
  m_visibleThreshold = 0.0f;
  m_visibleThresholdLoose = 0.0f;
  
  //m_tau = max(option.m_minImageNum * 2, min(m_num, 5));
  m_tau = min(option.m_minImageNum * 2, m_num);
  
  m_depth = 0;
  
  // set target images and other images
  m_bindexes = option.m_bindexes;
  m_visdata = option.m_visdata;
  m_visdata2 = option.m_visdata2;
  
  //----------------------------------------------------------------------
  pthread_rwlock_init(&m_lock, NULL);
  m_imageLocks.resize(m_num);
  m_countLocks.resize(m_num);
  for (int image = 0; image < m_num; ++image) {
    pthread_rwlock_init(&m_imageLocks[image], NULL);
    pthread_rwlock_init(&m_countLocks[image], NULL);
  }
  // We set m_level + 3, to use multi-resolutional texture grabbing
  m_pss.init(m_images, m_prefix, m_level + 3, m_wsize, 1);

  if (m_setEdge != 0.0f)
    m_pss.setEdge(m_setEdge);
  m_pss.setDistances();

  // Detect features if not yet done
  CdetectFeatures df;
  const int fcsize = 16;
  df.run(m_pss, m_num, fcsize, m_level, m_CPU);  
  
  // Initialize each core member. m_po should be first
  m_pos.init();
  m_seed.init(df.m_points);
  m_expand.init();
  m_filter.init();
  m_optim.init();
  //----------------------------------------------------------------------
  // Init thresholds
  m_angleThreshold0 = 60.0f * M_PI / 180.0f;
  m_angleThreshold1 = 60.0f * M_PI / 180.0f;

  m_countThreshold0 = 2;
  m_countThreshold1 = 4;
  m_countThreshold2 = 2;

  m_neighborThreshold = 0.5f;
  m_neighborThreshold1 = 1.0f;
  
  m_neighborThreshold2 = 1.0f;

  m_maxAngleThreshold = option.m_maxAngleThreshold;

  m_nccThresholdBefore = m_nccThreshold - 0.3f;
  
  m_quadThreshold = option.m_quadThreshold;
  
  m_epThreshold = 2.0f;
}

int CfindMatch::insideBimages(const Vec4f& coord) const {
  for (int i = 0; i < (int)m_bindexes.size(); ++i) {
    const int index = m_bindexes[i];
    const Vec3f icoord = m_pss.project(index, coord, m_level);
    if (icoord[0] < 0.0 || m_pss.getWidth(index, m_level) - 1 < icoord[0] ||
        icoord[1] < 0.0 || m_pss.getHeight(index, m_level) - 1 < icoord[1])
      return 0;
  }
  return 1;
}

int CfindMatch::isNeighbor(const Patch::Cpatch& lhs, const Patch::Cpatch& rhs,
                           const float neighborThreshold) const {  
  const float hunit = (m_optim.getUnit(lhs.m_images[0], lhs.m_coord) +
                       m_optim.getUnit(rhs.m_images[0], rhs.m_coord)) / 2.0
    * m_csize;
  return isNeighbor(lhs, rhs, hunit, neighborThreshold);
}

int CfindMatch::isNeighbor(const Patch::Cpatch& lhs, const Patch::Cpatch& rhs,
                           const float hunit, const float neighborThreshold) const {
  if (lhs.m_normal * rhs.m_normal < cos(120.0 * M_PI / 180.0))
    return 0;
  const Vec4f diff = rhs.m_coord - lhs.m_coord;
  
  const float vunit = lhs.m_dscale + rhs.m_dscale;
  
  const float f0 = lhs.m_normal * diff;
  const float f1 = rhs.m_normal * diff;   
  float ftmp = (fabs(f0) + fabs(f1)) / 2.0;
  ftmp /= vunit;

  // this may loosen the isneighbor testing. need to tighten (decrease) threshold?  
  const float hsize = norm(2 * diff - lhs.m_normal * f0 - rhs.m_normal * f1) / 2.0 / hunit;
  if (1.0 < hsize)
    ftmp /= min(2.0f, hsize);
  
  if (ftmp < neighborThreshold)
    return 1;
  else
    return 0;
}

int CfindMatch::isNeighborRadius(const Patch::Cpatch& lhs,
                                 const Patch::Cpatch& rhs,
                                 const float hunit,
                                 const float neighborThreshold,
                                 const float radius) const {
  if (lhs.m_normal * rhs.m_normal < cos(120.0 * M_PI / 180.0))
    return 0;
  const Vec4f diff = rhs.m_coord - lhs.m_coord;
  
  const float vunit = lhs.m_dscale + rhs.m_dscale;
  
  const float f0 = lhs.m_normal * diff;
  const float f1 = rhs.m_normal * diff;   
  float ftmp = (fabs(f0) + fabs(f1)) / 2.0;
  ftmp /= vunit;

  // this may loosen the isneighbor testing. need to tighten (decrease) threshold?  
  const float hsize = norm(2 * diff - lhs.m_normal * f0 - rhs.m_normal * f1) / 2.0 / hunit;

  // radius check
  if (radius / hunit < hsize)
    return 0;
  
  if (1.0 < hsize)
    ftmp /= min(2.0f, hsize);
  
  if (ftmp < neighborThreshold)
    return 1;
  else
    return 0;
}

void CfindMatch::run(void) {
  struct timeval tv;
  gettimeofday(&tv, NULL); 
  time_t curtime = tv.tv_sec;
  
  //----------------------------------------------------------------------
  // Seed generation
  m_seed.run();
  m_seed.clear();
  
  ++m_depth;
  m_pos.collectPatches();

  const int TIME = 3;
  int time;
  time_t write_time[TIME]; 
//----- cktang
  for (time = 0; time < TIME; time++) {
    m_expand.tvrun();
    {
       cerr << "write patches before filtering " << endl; 
       char buffer[1024];
       sprintf (buffer, "%s_before_filter", m_prefix.c_str());
       writePatches (m_prefix.c_str(),time); 
    }
    m_filter.run();
    gettimeofday (&tv, NULL); write_time[time] = tv.tv_sec;
    writePatches (m_prefix.c_str(),time);  // disk access time should be ignored
    gettimeofday (&tv, NULL);
    write_time[time] = tv.tv_sec - write_time[time]; 
    updateThreshold();
    cout << "STATUS: ";
    for (int i = 0; i < (int)m_optim.m_status.size(); ++i) {
      cout << m_optim.m_status[i] << ' ';
      if (i % 10 == 9)
        cout << endl;
    }
    cout << endl;
    ++m_depth;
  }
  gettimeofday(&tv, NULL); 
  time_t totalwt = 0;
  for (time = 0; time < TIME; time++)
    totalwt += write_time[time];

  cerr << "---- Total: " << tv.tv_sec - curtime - totalwt << " secs ----" << endl;
  return; 
//----- cktang
  
  //----------------------------------------------------------------------
  // Expansion
  for (int t = 0; t < TIME; ++t) {
    m_expand.run();

    m_filter.run();
        
    updateThreshold();

    cout << "STATUS: ";
    for (int i = 0; i < (int)m_optim.m_status.size(); ++i) {
      cout << m_optim.m_status[i] << ' ';
      if (i % 10 == 9)
        cout << endl;
    }
    cout << endl;
    
    ++m_depth;
  }
  cerr << "---- Total: " << tv.tv_sec - curtime << " secs ----" << endl;
}

void CfindMatch::write(const std::string prefix) {
  m_pos.writePatches2(prefix);
}

void CfindMatch::writePatches(const std::string prefix, int iter) {
  m_pos.writePatches3(prefix,iter);
}

void CfindMatch::sparseTV (int numPt, ANNkd_tree *annTree, double*** K, int iter, int m_iDim )
{
        int i, j, k;

	cerr << "SPARSE VOTING with sigma = " << m_fSigmaScale << endl;
        // ANNpointArray x = annTree->thePoints();
        ANNpointArray x = m_annTree->thePoints();

        int* indexArray = new int[ numPt ];  
        double* distArray = new double [ numPt ];
        double** newK = allocate2DDouble( m_iDim, m_iDim );
        double** tempS = allocate2DDouble( m_iDim, m_iDim );
        double** M = allocate2DDouble( m_iDim, m_iDim );
        double** eigVec = allocate2DDouble( m_iDim, m_iDim );
        double*  eigVal = new double[ m_iDim ];
        double*  newV = new double[ m_iDim ];
        double*** Kp = new double**[ numPt ];
        for( i = 0; i < numPt; i++ )
                Kp[i] = allocate2DDouble( m_iDim, m_iDim );

        double radius2 = - m_fSigmaScale * log( m_fEpsBoundary );

        for( k = 0; k < iter; k++ )
        {
                // Estimate K[i]
                int numNeighbor = 0;
                for( i = 0; i < numPt; i++ )
                {
                        numNeighbor = numPt;
			numNeighbor = m_annTree->annkFRSearch( (ANNpoint) x[i], (ANNdist) radius2, numNeighbor, (ANNidxArray) indexArray, (ANNdistArray) distArray );

			if (i==0) cerr << "No of neighbors = " << numNeighbor << endl;

                        zeroMat( m_iDim, m_iDim, newK );

                        for( j = 0; j < numNeighbor; j++ )
                        {
                                int index = indexArray[j];

                                // find S here
                                zeroMat( m_iDim, m_iDim, tempS );

                                getTensor( x[i], x[index], K[index], tempS, m_fSigmaScale, m_iDim );

//                                if( isZeroMat( m_iDim, m_iDim, tempS ) ) printf( "zeroMat\n" );

                                addMat( m_iDim, m_iDim, newK, tempS, newK );

                        }

                        for( j = 0; j < m_iDim; j++ )
                                newK[j][j] += 0.01;
                        
                        copyMatrix( newK, Kp[i], m_iDim, m_iDim );
                }

                for( i = 0; i < numPt; i++ )
                        copyMatrix( Kp[i], K[i], m_iDim, m_iDim );
        }

        // record maximum surface saliency and then normalize K

        double maxsal = 0.0;
        for( i = 0; i < numPt; i++ )
        {
		double sal; 
                eig_sys( m_iDim, K[i], eigVec, eigVal );
                sal = eigVal[0]-eigVal[1];
		maxsal = (sal > maxsal) ? sal : maxsal;
                matConstMultiply( m_iDim, m_iDim, K[i], 1.0 / eigVal[0], K[i] );
        }
        
  	m_maxsal = maxsal;
        cerr << "maximum saliency = " << m_maxsal << endl;
        annClose();

        SAFEDELARR( indexArray );
        SAFEDELARR( distArray );
        safeDelete2DDouble( m_iDim, newK );
        safeDelete2DDouble( m_iDim, tempS );
        safeDelete2DDouble( m_iDim, M );
        safeDelete2DDouble( m_iDim, eigVec );
        SAFEDELARR( eigVal );

        if( Kp != NULL )
        {
                for( i = 0; i < numPt; i++ )
                        safeDelete2DDouble( m_iDim, Kp[i] );
                SAFEDELARR( Kp );
        }
	cerr << "SPARSE VOTING ends... " << endl;
}


Vec4f CfindMatch::getNormalfromTV (Cpatch& patch, const int numPt, ANNkd_tree *annTree, double*** K)
{
  
/* TODO:
   1) get NumPts neighbors of patch.m_coord using annTree->annkFRSearch
   2) collect votes cast by NumPts neighors using getTensor
   3) after vote collection, return the eigenvector corr to the largest eigenvalue 
*/

  // ANNpointArray x = annTree->thePoints();
  ANNpointArray x = m_annTree->thePoints();
  int* indexArray = new int[ numPt ];
  double* distArray = new double [ numPt ];
  double** newK = allocate2DDouble( 3, 3 );
  double** tempS = allocate2DDouble( 3, 3 );
  double** eigVec = allocate2DDouble( 3, 3 );
  double*  eigVal = new double[ 3 ];

  int i, j;
  // int numPts = annTree->nPoints();
  int numPts = m_annTree->nPoints();

  double annpt[3];
  for (i=0;i<3;i++) 
    annpt[i] = patch.m_coord[i];

  double radius2 = - m_fSigmaScale * log( m_fEpsBoundary );

  // int numNeighbors = annTree->annkFRSearch( (ANNpoint) annpt, (ANNdist) radius2, numPts, 
  //   (ANNidxArray) indexArray, (ANNdistArray) distArray ); 
  int numNeighbors = m_annTree->annkFRSearch( (ANNpoint) annpt, (ANNdist) radius2, numPts, 
     (ANNidxArray) indexArray, (ANNdistArray) distArray ); 
 
  zeroMat( 3, 3, newK );

  for( j = 0; j < numNeighbors; j++ ) 
  {
    int index = indexArray[j];
    // find S here
    zeroMat( 3, 3, tempS );

    getTensor( annpt, x[index], K[index], tempS, m_fSigmaScale, 3 );
    addMat( 3, 3, newK, tempS, newK );

   }

  for( j = 0; j < 3; j++ )
    newK[j][j] += 0.01;

  eig_sys( 3, newK, eigVec, eigVal );
  patch.m_sal = eigVal[0] - eigVal[1];
  matConstMultiply( 3, 3, newK, 1.0 / eigVal[0], newK );

  annClose();


  // Vec4f v(sal*eigVec[0][0],sal*eigVec[0][1],sal*eigVec[0][2], 1.0); 
  Vec4f v(eigVec[0][0],eigVec[0][1],eigVec[0][2], 1.0); 

  SAFEDELARR( indexArray );
  SAFEDELARR( distArray );
  safeDelete2DDouble( 3, eigVec );
  SAFEDELARR( eigVal );
  safeDelete2DDouble( 3, newK );
  safeDelete2DDouble( 3, tempS );

  return v;

}

void CfindMatch::getTensor( double* x1, double* x2, double** K2, double** S, double sigma, int m_iDim )
{

        assert( x1 != NULL );
        assert( x2 != NULL );
        assert( K2 != NULL );
        assert( S  != NULL );

        int i, j;

        if( R == NULL ) R = allocate2DDouble( m_iDim, m_iDim );
        if( Rp == NULL ) Rp = allocate2DDouble( m_iDim, m_iDim );
        if( outr == NULL ) outr = allocate2DDouble( m_iDim, m_iDim );
        if( r == NULL ) r = new double[ m_iDim ];
        if( I == NULL ) I = allocate2DDouble( m_iDim, m_iDim );
        eye( m_iDim, I );
	
        for( i = 0; i < m_iDim; i++ )
                r[i] = x1[i] - x2[i];

        double length_r2 = norm2( m_iDim, r );
        double length_r  = sqrt( length_r2 );

        if( !nearzero( length_r ) )
                for( i = 0; i < m_iDim; i++ )
                        r[i] = r[i] / length_r;

        // outer product of r
        outerProduct( m_iDim, r, m_iDim, r, outr );

        for( i = 0; i < m_iDim; i++ )
                for( j = 0; j < m_iDim; j++ )
                        outr[i][j] = 2.0 * outr[i][j];

        // find R
        subMat( m_iDim, m_iDim, I, outr, R );

        for( i = 0; i < m_iDim; i++ )
                for( j = 0; j < m_iDim; j++ )
                        outr[i][j] = outr[i][j] / 4.0;

        // find Rp
        subMat( m_iDim, m_iDim, I, outr, Rp );
        multiMat( m_iDim, m_iDim, m_iDim, Rp, R, outr );
        copyMatrix( outr, Rp, m_iDim, m_iDim );

        //find S
        multiMat( m_iDim, m_iDim, m_iDim, R, K2, outr );
        multiMat( m_iDim, m_iDim, m_iDim, outr, Rp, S );
        
        double c = exp( - fabs( length_r2 ) / sigma );

        // scale the S by distance
        for( i = 0; i < m_iDim; i++ )
                for( j = 0; j < m_iDim; j++ )
                        S[i][j] = c * S[i][j];
}

