//=======================================================================
// main.cpp : The implementation of Markov Random Field Tensor Voting on 
//            filtering, corresponds to Section 6 of the paper "Quasi-Dense 3D 
//            Reconstruction using Tensor-based Multiview Stereo", CVPR'10.
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//======================================================================= 

#include "StdAfx.h"
#include "cftv.h"
#include "mrftv.h"
#include "eig.h"
#include "utility.h"
#include "quicksort.h"

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>

#include <time.h>

#include <vector>
#include <list>
using namespace std;

int main( int argc, char* argv[] )
{

	double  sigma = 0.1;
	double  lambda = 1.0;
	double  epsBoundary = 0.075;
	int     maxIter = 10000000;
	double  lineScale = 1.0;
	char*    tensorFile = NULL;
	bool    outputTensor = false;
	double  noiseThreshold = 0.1;
	double  sor = 1.0;

	if( argc <= 2 )
	{
		printf( "\n" );
		printf( "=======================================================================\n");
		printf( "TV Filtering v1.0\n" );
		printf( "Copyright (C) 2009 T-P WU\n" );
		printf( "-----------------------------------------------------------------------\n");
		printf( "Usage: %s <input file> <output file> [-s|l|eps|i|ts|nt|ot <value>]\n", argv[0]);
		printf( "I/O format: bundler or viewworld (point)\n" );
		printf( "Description: Perform MRF Tensor Voting to remove outliers.\n");
		printf( "    -s:  Scale of analysis;       value -- (0, 1]      in (real)\n" );
		printf( "    -l:  Neighborhood weighting;  value -- [0, inf)    in (real)\n" );
		printf( "  -eps:  Gaussian tail cutoff;    value -- [0, 1)      in (real)\n" );
		printf( "    -i:  Number of iteration;     value -- [1, inf)    in (int)\n" );
		printf( "  -sor:  SOR weight;              value -- [1, 2)      in (real)\n");
		printf( "   -ts:  Scale of output tensors; value -- (0, inf)    in (real)\n" );
		printf( "   -nt:  Noise threshold;         value -- [0, 1]      in (real)\n" );
		printf( "   -ot:  Output tensor file;      value -- {full path} in (string)\n" );
		printf( "Default values:\n" );
		printf( "     s = %0.3f; l = %0.3f; eps = %0.3f; i = %d;\n", sigma, lambda, epsBoundary, maxIter );
		printf( "     sor = %0.3f; ts = %0.3f; nt = %0.3f;\n", sor, lineScale, noiseThreshold );
		printf( "A note on SOR: the value of sor will be estimated automatically\n" );
		printf( "               when (1 < sor) or (sor >= 2).\n" );
		printf( "=======================================================================\n");
		printf( "\n" );
		return 0;
	}
	
	int i, j;
	
	clock_t startTime, endTime;
	
	double**         point = NULL;
	double**         point2 = NULL;
	double**         pointRGB = NULL;
	list<int>*       visibleCam = NULL;
	list<int>*       siftKey = NULL;
	list<double>*    ix = NULL;
	list<double>*    iy = NULL;
	double*          focalLength = NULL;
	double**         radial = NULL;
	double***        R = NULL;
	double**         t = NULL;
	int              numPoint = 0;
	int              numCamera = 0;

	bool*            cleanPoint = NULL;

	// read file extension
	char path[1024];
	strcpy( path, argv[1] );
	char* extension = fileExtension(path);	

	// read file
	if( strcmp( extension, "vw" ) == 0 )
	{
		// read vw file
		printf( "[TV Filtering] Read viewworld file: %s \n", argv[1] );
		startTime = clock();
		readVW3DPoint( point, numPoint, argv[1] );
		endTime = clock();
		printf("[TV Filtering] Read viewworld file took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));
	}
	else
	{
		// read bundle file
		printf( "[TV Filtering] Read bundle file: %s \n", argv[1] );
		startTime = clock();
		readBundle( argv[1], point, pointRGB, visibleCam, siftKey, ix, iy, numPoint, focalLength, radial, R, t, numCamera );
		endTime = clock();
		printf("[TV Filtering] Read bundle file took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));
	}

	double*** K = new double**[numPoint];
	for( i = 0; i < numPoint; i++ )
	{
		K[i] = allocate2DDouble( 3, 3 );
		eye( 3, K[i] );
	}
	
	// parse the input arguments
	for( i = 3; i < argc; i = i + 2 )
	{
		if( strcmp(argv[i], "-s" ) == 0 )    sigma = atof( argv[i+1] );
		if( strcmp(argv[i], "-l" ) == 0 )    lambda = atof( argv[i+1] );
		if( strcmp(argv[i], "-eps" ) == 0 )  epsBoundary = atof( argv[i+1] );
		if( strcmp(argv[i], "-i" ) == 0 )    maxIter = atoi( argv[i+1] );
		if( strcmp(argv[i], "-sor" ) == 0 )  sor = atof( argv[i+1] );
		if( strcmp(argv[i], "-ts" ) == 0 )   lineScale = atof( argv[i+1] );
		if( strcmp(argv[i], "-nt") == 0 )    noiseThreshold = atof( argv[i+1] );
		if( strcmp(argv[i], "-ot" ) == 0 )   { outputTensor = true; tensorFile = argv[i+1]; }
	}

	if( !( (sor >= 1.0) && (sor < 2.0 - EPLSION) ) )
		printf( "[TV Filtering] SOR parameter update enabled.\n" );

	// find the max and min coordinate in 3D to scale sigma
	double maxCoord[3] = { -MYINFINITE, -MYINFINITE, -MYINFINITE };
	double minCoord[3] = { MYINFINITE, MYINFINITE, MYINFINITE };
	double dimLength[3] = {0.0, 0.0, 0.0}; 
	for( i = 0; i < numPoint; i++ )
		for( j = 0; j < 3; j++ )
		{
			if( point[i][j] < minCoord[j] ) minCoord[j] = point[i][j];
			if( point[i][j] > maxCoord[j] ) maxCoord[j] = point[i][j];
		}
	for( j = 0; j < 3; j++ )
		dimLength[j] = maxCoord[j] - minCoord[j];
	int minAxisIndex = 0;
	for( j = 1; j < 3; j++ )
		if( dimLength[minAxisIndex] > dimLength[j] ) minAxisIndex = j;
	sigma = sigma * dimLength[minAxisIndex];

	printf( "[TV Filtering] Adjusted sigma = %f\n", sigma );

	// Perform standard Tensor Voting with Closed-from TV
	// Here, I used this method to obtain \tilde{K} (Eqn 11 of (*)) 
	// which shows you how standard tensor voting (with CFTV) is called.
	printf( "[TV Filtering] Start standard TV with CFTV .... \n" );
	startTime = clock();
	tensorVoting( K, point, numPoint, 3, sigma * sigma, 2, epsBoundary );
	endTime = clock();
	printf( "[TV Filtering] Standard TV Done\n" );
	printf( "[TV Filtering] Standard TV took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC) ); 

	// normalize the resultant tensors (It is not necessary. just want to make the scale more tractable)
	normalizeTensors( K, numPoint, 3 );	

	// perform TV-MRF to obtain diffused tensors
	printf( "[TV Filtering] Start TV-MRF .... \n" );
	startTime = clock();
	int iteration = votingMRF( K, point, numPoint, 3, sigma * sigma, lambda, epsBoundary, sor, maxIter );
	endTime = clock();
	printf( "[TV Filtering] TV-MRF Done\n" );
	printf("[TV Filtering] TV-MRF took %0.3fs with %d iterations\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC), iteration);


	double** endPt1 = NULL;
	double** endPt2 = NULL;
	int      numLineSeg = 0;

	// output tensors (as principle axis)
	if( outputTensor )
	{
		printf( "[TV Filtering] Output tensors .... ");
		startTime = clock();
		tensors2Axis( endPt1, endPt2, numLineSeg, K, point, numPoint, 3, lineScale);
		saveVWLineSeg( endPt1, endPt2, numLineSeg, tensorFile );
		endTime = clock();
		printf( "Done\n");
		printf("[TV Filtering] Output vw file took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));
	}

	// find the clean points
	printf( "[TV Filtering] Filter out noise .... ");
	startTime = clock();
	filterNoise( cleanPoint, K, numPoint, 3, noiseThreshold );
	endTime = clock();
	printf( "Done\n");
	printf("[TV Filtering] Filter out noise took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));

	//save clean data
	if( strcmp( extension, "vw" ) == 0 )
	{
		// save the clean viewworld file
		printf( "[TV Filtering] Save viewworld file: %s \n", argv[1] );
		startTime = clock();
		saveVW3DPoint( point, numPoint, argv[2], cleanPoint );
		endTime = clock();
		printf("[TV Filtering] Save viewworld file took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));
	}
	else
	{
		// save the clean bundle file
		printf( "[TV Filtering] Save bundle file: %s \n", argv[1] );
		startTime = clock();
		saveBundle( argv[2], point, pointRGB, visibleCam, siftKey, ix, iy, numPoint, focalLength, radial, R, t, numCamera, cleanPoint );
		endTime = clock();
		printf("[TV Filtering] Save bundle file took %0.3fs\n", (endTime - startTime) / ((double) CLOCKS_PER_SEC));
	}

	printf( "[TV Filtering] Done.\n");

	if( K != NULL )
	{
		for( i = 0; i < numPoint; i++ )
			safeDelete2DDouble( 3, K[i] );
		SAFEDELARR( K );
	}
	safeDelete2DDouble( numLineSeg, endPt1 );
	safeDelete2DDouble( numLineSeg, endPt2 );

	SAFEDELARR( extension );
	safeDelete2DDouble( numPoint, point );
	safeDelete2DDouble( numPoint, point2 );
	safeDelete2DDouble( numPoint, pointRGB );
	SAFEDELARR( visibleCam );
	SAFEDELARR( siftKey );
	SAFEDELARR( ix );
	SAFEDELARR( iy );
	SAFEDELARR( focalLength );
	safeDelete2DDouble( numCamera, radial );
	safeDelete2DDouble( numCamera, t );
	if( R != NULL )
	{
		for( i = 0; i < numCamera; i++ )
			safeDelete2DDouble( 3, R[i] );
		SAFEDELARR( R );
	}
	SAFEDELARR( cleanPoint );

	return 1;
}


