//=======================================================================
// utility.cpp : The implementation of utility.h
//
// Copyright (c) 2009 Tai-Pang Wu (pang@cse.ust.hk) and HKUST x CUHK
// 
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//======================================================================= 


#include "StdAfx.h"
#include "utility.h"
#include "eig.h"

int readVW3DPoint( double**& point, int& numPoint, const char* filename )
{
	FILE* fid = fopen( filename, "r" );
	
	if( fid == NULL )
	{
		printf( "Error: Cannot load vw file [%s]!\n", filename );
		return 0;
	}

	list<double> xList;
	list<double> yList;
	list<double> zList;
	
	double x, y, z;
	int d;
	numPoint = 0; 

	while( fscanf(fid, "%lf %lf %lf %d\n", &x, &y, &z, &d) == 4 )
	{
		if( d == 2 )
		{
			xList.push_back( x );
			yList.push_back( y );
			zList.push_back( z );
			numPoint++;
		}
	}

	fclose( fid );

	if( point != NULL ) safeDelete2DDouble( numPoint, point );
	point = allocate2DDouble( numPoint, 3 );

	list<double>::iterator itx = xList.begin();
	list<double>::iterator ity = yList.begin();
	list<double>::iterator itz = zList.begin();

	int i = 0;
	while( itx != xList.end() )
	{
		point[i][0] = *itx;
		point[i][1] = *ity;
		point[i][2] = *itz;

		itx++;
		ity++;
		itz++;
		i++;
	}

	xList.clear();
	yList.clear();
	zList.clear();

	return 1;
}

int saveVW3DPoint( double** point, int numPoint, const char* filename, bool* cleanPoint )
{
	if( point == NULL ) return 0;
	
	FILE* fid = fopen( filename, "w" );
	
	if( fid == NULL )
	{
		printf( "Error: Cannot save vw file [%s]!\n", filename );
		return 0;
	}
	
	int i = 0;
	for( i = 0; i < numPoint; i++ )
	{
		if( cleanPoint != NULL && !cleanPoint[i] ) continue;
		
		fprintf( fid, "%f %f %f 2\n", point[i][0], point[i][1], point[i][2] );
	}
	
	fclose( fid );
	
	return 1;
}

int saveVWLineSeg( double** endPt1, double** endPt2, int numLineSeg, const char* filename )
{
	if( endPt1 == NULL ) return 0;
	if( endPt2 == NULL ) return 0;
	if( numLineSeg <= 0 ) return 0;

	FILE* fid = fopen( filename, "w" );
	if( fid == NULL )
	{
		printf( "Error: Cannot save vw file [%s]!\n", filename );
		return 0;
	}

	int i;
	for( i = 0; i < numLineSeg; i = i + 1 )
	{
		fprintf( fid, "%f %f %f 0\n", endPt1[i][0], endPt1[i][1], endPt1[i][2] );
		fprintf( fid, "%f %f %f 1\n", endPt2[i][0], endPt2[i][1], endPt2[i][2] );
	}

	fclose( fid );

	return 1;
}

int tensors2Axis( double**& endPt1, double**& endPt2, int& numLineSeg, double*** K, double** x, int numPoint, int dim, double axisScale)
{
	if( K == NULL ) return 0;
	if( x == NULL ) return 0;
	if( numPoint <= 0 ) return 0;
	if( dim <= 0 ) return 0;

	if( endPt1 != NULL ) safeDelete2DDouble( numLineSeg, endPt1 );
	if( endPt2 != NULL ) safeDelete2DDouble( numLineSeg, endPt2 );

	numLineSeg = numPoint * dim;
	endPt1 = allocate2DDouble( numLineSeg, dim );
	endPt2 = allocate2DDouble( numLineSeg, dim );

	double** eigVec = allocate2DDouble( dim, dim );
	double*  eigVal = new double[ dim ];

	int i, j, k;
	
	for( i = 0; i < numPoint; i++ )
	{
		eig_sys( dim, K[i], eigVec, eigVal );

		// [0]:largest, [dim-1]:smallest		
		for( j = 0; j < dim; j++ )
		{
			for( k = 0; k < dim; k++ )
			{
				endPt1[ dim * i + j ][k] = x[i][k] + eigVec[j][k] * eigVal[j] * axisScale;
				endPt2[ dim * i + j ][k] = x[i][k] - eigVec[j][k] * eigVal[j] * axisScale;
			}
		}

	}

	safeDelete2DDouble( dim, eigVec );
	SAFEDELARR( eigVal );

	return 1;
}


int readBundle( const char* filename, double**& point, double**& pointRGB, 
			   list<int>*& visibleCam, list<int>*& siftKey, list<double>*& ix, list<double>*& iy, int& numPoint, 
			   double*& focalLength, double**& radial, double***& R, double**& t, int& numCamera)
{
	int i, j;

	FILE* fp = fopen( filename, "r" );
	if( fp == NULL )
	{
		printf( "[TensorVoting MRF] Error: Cannot read bundler file [%s]!\n", filename );
		return 0;
	}

	if( point != NULL ) safeDelete2DDouble( numPoint, point );
	if( pointRGB != NULL ) safeDelete2DDouble( numPoint, pointRGB );
	if( focalLength != NULL ) SAFEDELARR( focalLength );
	if( radial != NULL ) safeDelete2DDouble( numCamera, radial );
	if( t != NULL ) safeDelete2DDouble( numCamera, t );
	if( R != NULL )
	{
		for( i = 0; i < numCamera; i++ )
			safeDelete2DDouble( numCamera, R[i] );
		SAFEDELARR( R );
	}
	if( visibleCam != NULL ) SAFEDELARR( visibleCam );
	if( siftKey != NULL ) SAFEDELARR( siftKey );
	if( ix != NULL ) SAFEDELARR( ix );
	if( iy != NULL ) SAFEDELARR( iy );

	// read header
	char buffer[1024];
	for( i = 0 ; i < 4; i++ )
		fscanf( fp, "%s ", buffer );

	// Read number of camera and points
	fscanf( fp, "%ld %ld", &numCamera, &numPoint );

	point = allocate2DDouble( numPoint, 3 );
	pointRGB = allocate2DDouble( numPoint, 3 );
	focalLength = new double[numCamera];
	radial = allocate2DDouble( numCamera, 2 );
	t = allocate2DDouble( numCamera, 3 );
	R = new double**[numCamera];
	for( i = 0; i < numCamera; i++ )
		R[i] = allocate2DDouble( 3, 3 );
	visibleCam = new list<int>[numPoint];
	siftKey = new list<int>[numPoint];
	ix  = new list<double>[numPoint];
	iy  = new list<double>[numPoint];

	// read camera
	for ( i = 0; i < numCamera; i++ )
	{
		fscanf( fp, "%lf %lf %lf\n", &focalLength[i], &radial[i][0], &radial[i][1] );
		fscanf( fp, "%lf %lf %lf\n", &R[i][0][0], &R[i][0][1], &R[i][0][2] );  
		fscanf( fp, "%lf %lf %lf\n", &R[i][1][0], &R[i][1][1], &R[i][1][2] );
		fscanf( fp, "%lf %lf %lf\n", &R[i][2][0], &R[i][2][1], &R[i][2][2] );
		fscanf( fp, "%lf %lf %lf\n", &t[i][0], &t[i][1], &t[i][2] ); 
	}

	// read point
	int r, g, b;
	for( i = 0; i < numPoint; i++ )
	{
		// Read 3D position
		fscanf( fp, "%lf %lf %lf\n", &point[i][0], &point[i][1], &point[i][2] );

		// Read RGB
		fscanf( fp, "%d %d %d\n", &r, &g, &b );
		pointRGB[i][0] = double( r );
		pointRGB[i][1] = double( g );
		pointRGB[i][2] = double( b );

		// Read view list
		int numCamVis, camIdx, key;
		double im_x, im_y;

		fscanf( fp, "%d ", &numCamVis );
		for ( j = 0; j < numCamVis ; j++ )
		{
			fscanf( fp, "%d %d %lf %lf", &camIdx, &key, &im_x, &im_y ); 
			visibleCam[i].push_back( camIdx );
			siftKey[i].push_back( key );
			ix[i].push_back( im_x );
			iy[i].push_back( im_y ); 
		}

	}

	fclose( fp );


	return 1;
}

int saveBundle( const char* filename, double** point, double** pointRGB, 
			   list<int>* visibleCam, list<int>* siftKey, list<double>* ix, list<double>* iy, int numPoint, 
			   double* focalLength, double** radial, double*** R, double** t, int numCamera,
			   bool* cleanPoint)
{
	if( (filename == NULL) || ( point == NULL ) || ( pointRGB == NULL ) || ( visibleCam == NULL ) 
		|| ( siftKey == NULL ) || ( ix == NULL ) || ( iy == NULL ) || ( focalLength == NULL )
		|| ( radial == NULL ) || ( R == NULL ) || ( t == NULL ) ) 
		return 0;

	int i, j;

	FILE* fid = fopen( filename, "w" );
	if( fid == NULL )
	{
		printf( "[TensorVoting MRF] Error: Cannot save bundler file [%s]!\n", filename );
		return 0;
	}

	// print header
	fprintf( fid, "# Bundle file v0.3\n" );

	int numCleanPoint = numPoint;
	if( cleanPoint != NULL )
	{
		numCleanPoint = 0;
		for( i = 0; i < numPoint; i++ )
			if( cleanPoint[i] ) numCleanPoint++;
	}

	// print number of camera and number of point
	fprintf( fid, "%d %d\n", numCamera, numCleanPoint );

	// print cameras
	for( i = 0; i < numCamera; i++ )
	{
		// print focal length and radial distortion coeffs
		fprintf( fid, "%f %f %f\n", focalLength[i], radial[i][0], radial[i][1] );

		// print rotation matrix
		for( j = 0; j < 3; j++ )
			fprintf( fid, "%f %f %f\n", R[i][j][0], R[i][j][1], R[i][j][2] );

		// printf translation 
		fprintf( fid, "%f %f %f\n", t[i][0], t[i][1], t[i][2] );
	}

	// print clean points
	for( i = 0; i < numPoint; i++ )
	{
		if( (cleanPoint != NULL) && (!cleanPoint[i]) )
			continue;

		//print 3D location
		fprintf( fid, "%f %f %f\n", point[i][0], point[i][1], point[i][2] );

		//print RGB
		fprintf( fid, "%d %d %d\n", int(pointRGB[i][0]), int(pointRGB[i][1]), int(pointRGB[i][2]) );

		//print visible list
		if( visibleCam[i].size() > 0 )
			fprintf( fid, "%d ", visibleCam[i].size() );

		list<int>::iterator itCam = visibleCam[i].begin();
		list<int>::iterator itKey = siftKey[i].begin();
		list<double>::iterator itx = ix[i].begin();
		list<double>::iterator ity = iy[i].begin();

		while( itCam != visibleCam[i].end() )
		{
			fprintf( fid, "%d %d %f %f ", *itCam, *itKey, *itx, *ity );

			itCam++;
			itKey++;
			itx++;
			ity++;
		}

		if( visibleCam[i].size() > 0 ) fprintf( fid, "\n");
	}

	fclose( fid );

	return 1;
}

int normalizeTensors( double*** K, int numPoint, int dim )
{
	if( K == NULL ) return 0;
	if( numPoint <= 0 ) return 0;
	if( dim < 2 ) return 0;

	int i;

	double** eigVec = allocate2DDouble( dim, dim );
	double*  eigVal = new double[ dim ];

	double maxEigenValue = -1;

	for( i = 0; i < numPoint; i++ )
	{
		eig_sys( dim, K[i], eigVec, eigVal );
		if( maxEigenValue < eigVal[0] ) maxEigenValue = eigVal[0];
	}

	double factor = 1.0 / maxEigenValue;
	for( i = 0; i < numPoint; i++ )
		matConstMultiply( dim, dim, K[i], factor, K[i] );
	
	safeDelete2DDouble( dim, eigVec );
	SAFEDELARR( eigVal );

	return 1;
}

int filterNoise( bool*& clean, double*** K, int numPoint, int dim, double threshold )
{
	if( K == NULL ) return 0;
	if( numPoint <= 0 ) return 0;
	if( dim < 2 ) return 0;

	int i;

	SAFEDELARR( clean );
	clean = new bool[numPoint];
	for( i = 0; i < numPoint; i++ )
		clean[i] = false;

	double* surfaceSaliency = new double[numPoint];
	memset( surfaceSaliency, 0, sizeof( double ) * numPoint );

	double minSaliency = MYINFINITE;
	double maxSaliency = -MYINFINITE;
	
	double** eigVec = allocate2DDouble( dim, dim );
	double*  eigVal = new double[ dim ];

	for( i = 0; i < numPoint; i++ )
	{
		eig_sys( dim, K[i], eigVec, eigVal );
		surfaceSaliency[i] = (eigVal[0] - eigVal[1]) * eigVal[0];
		
		if( i == 0 )
		{
			minSaliency = surfaceSaliency[i];
			maxSaliency = surfaceSaliency[i];
		}

		if( minSaliency > surfaceSaliency[i] ) minSaliency = surfaceSaliency[i];
		if( maxSaliency < surfaceSaliency[i] ) maxSaliency = surfaceSaliency[i];
	}

	if( !nearzero( maxSaliency ) )
		for( i = 0; i < numPoint; i++ )
			surfaceSaliency[i] = surfaceSaliency[i] / maxSaliency;
	
	for( i = 0; i < numPoint; i++ )
		if( surfaceSaliency[i] > threshold ) 
			clean[i] = true;
	
	safeDelete2DDouble( dim, eigVec );
	SAFEDELARR( eigVal );
	SAFEDELARR( surfaceSaliency );

	return 1;
}


